unit NQCSerial;

interface

uses
  Windows;

const
  kPStream_NeverTimeout = -1;
  kPSerial_DataMask	 = $03;
  kPSerial_DataShift = 0;
  kPSerial_Data8     = (0 shl kPSerial_DataShift);
  kPSerial_Data7     = (1 shl kPSerial_DataShift);
  kPSerial_Data6     = (2 shl kPSerial_DataShift);
  kPSerial_Data5     = (3 shl kPSerial_DataShift);

  kPSerial_ParityMask	 = $0c;
  kPSerial_ParityShift = 2;
  kPSerial_ParityNone  = (0 shl kPSerial_ParityShift);
  kPSerial_ParityOdd   = (1 shl kPSerial_ParityShift);
  kPSerial_ParityEven	 = (2 shl kPSerial_ParityShift);

  kPSerial_StopMask	 = $30;
  kPSerial_StopShift = 4;
  kPSerial_Stop1     = (0 shl kPSerial_StopShift);
  kPSerial_Stop1_5   = (1 shl kPSerial_StopShift);
  kPSerial_Stop2     = (2 shl kPSerial_StopShift);

  kDefaultSpeed = 9600;
  kInputBufferSize = 16384;
  kOutputBufferSize = 16384;

  fBinary           = $0001; // Binary mode if set
  fParity           = $0002; // When set, parity checking is enabled.
  fOutxCtsFlow      = $0004; // No data sent unless CTS is high.
  fOutxDsrFlow      = $0008; // No data sent unless DSR is high.
  fDtrControlEnable = $0010; // 1 = 10 (enabled), 2 = 20 (handshake)
  fDtrControlHand   = $0020; // DTR_CONTROL_DISABLE, DTR_CONTROL_ENABLE, DTR_CONTROL_HANDSHAKE
  fDsrSensitivity   = $0040; // Unless DSR is high, all bytes ignored.
  fTxContinueOnXOff = $0080; // Can continue sending data, even when waiting on an XON character to be set. If not set, cannot send data until XONLim is reached.
  fOutX             = $0100; // XON/XOFF flow control enabled for sending.
  fInX              = $0200; // XON/XOFF flow control enabled for receiving.
  fErrorChar        = $0400; // If a parity error is detected, the error will be replaced with this character.
  fNull             = $0800; // Strip off the null characters.
  fRtsControlEnable = $1000; // 1 = 1000 (enabled), 2 = 2000 (handshake)
  fRtsControlHand   = $2000; // RTS_CONTROL_DISABLE, RTS_CONTROL_ENABLE, RTS_CONTROL_HANDSHAKE
  fAbortOnError     = $4000; // abort on error

type
  TNQCStream = class
  protected
    fOpen : Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Close; virtual;
    function Read(Ptr : PByte; count : LongInt) : LongInt; virtual; abstract;
    function Write(Ptr : PByte; count : LongInt) : LongInt; virtual; abstract;
    procedure FlushWrite; virtual;
    function SetTimeout(timeout_ms : LongInt = kPStream_NeverTimeout) : Boolean; virtual;
    function ReadLine(data : PChar; max : Integer) : Boolean; virtual;
    function SetBlocking(bBlock : Boolean) : Boolean;
  end;

  TNQCSerial = class(TNQCStream)
  protected
    fFile : THandle;
  public
    function Open(Name : string) : Boolean; virtual;
    procedure Close; override;
    function Read(Ptr : PByte; count : LongInt) : LongInt; override;
    function Write(Ptr : PByte; count : LongInt) : LongInt; override;
    procedure FlushWrite; override;
    function SetTimeout(timeout_ms : LongInt = kPStream_NeverTimeout) : Boolean; override;
    function SetSpeed(speed : integer; opts : integer = 0) : Boolean; virtual;
    function SetDTR(bDTR : Boolean) : Boolean; virtual;
    function SetRTS(bRTS : Boolean) : Boolean; virtual;
  end;

  TNQCUsbTower = class(TNQCSerial)
  private
    fTimeoutValue : LongInt;
  public
    function Open(Name : string) : Boolean; override;
    function Read(Ptr : PByte; count : LongInt) : LongInt; override;
    function SetTimeout(timeout_ms : LongInt = kPStream_NeverTimeout) : Boolean; override;
    function SetSpeed(speed : integer; opts : integer = 0) : Boolean; override;
    function SetDTR(bDTR : Boolean) : Boolean; override;
    function SetRTS(bRTS : Boolean) : Boolean; override;
  end;

implementation

{ TNQCStream }

procedure TNQCStream.Close;
begin
  fOpen := false;
end;

constructor TNQCStream.Create;
begin
  inherited Create;
  fOpen := false;
end;

destructor TNQCStream.Destroy;
begin
  Close;
  inherited Destroy;
end;

procedure TNQCStream.FlushWrite;
begin
  // do nothing in base class
end;

function TNQCStream.ReadLine(data: PChar; max: Integer): Boolean;
begin
  result := false;
	while(max > 1) do begin
		if (Read(PByte(data), 1) <> 1) then
      break;
		dec(max);
		if ( data^ = #13) then
		begin
			inc(data);
			result := true;
			break;
		end;
		inc(data);
	end;
	data^ := #0;
end;

function TNQCStream.SetBlocking(bBlock: Boolean): Boolean;
begin
  if bBlock then
    result :=  SetTimeout(kPStream_NeverTimeout)
  else
    result :=  SetTimeout(0);
end;

function TNQCStream.SetTimeout(timeout_ms: Integer): Boolean;
begin
	result := (timeout_ms = kPStream_NeverTimeout);
end;

{ TNQCSerial }

procedure TNQCSerial.Close;
begin
  if not fOpen then exit;
  CloseHandle(fFile);
  fOpen := False;
end;

procedure TNQCSerial.FlushWrite;
begin
  FlushFileBuffers(fFile);
end;

function TNQCSerial.Open(Name: string): Boolean;
begin
  Result := false;
  if fOpen then exit;

  fFile := CreateFile(PChar(Name), GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING, 0, 0);

  if fFile = INVALID_HANDLE_VALUE then exit;

  if not SetSpeed(kDefaultSpeed) then exit;
  SetBlocking(True);
  fOpen := True;
  Result := True;
end;

function TNQCSerial.Read(Ptr : PByte; count: Integer): LongInt;
var
  actual, Errors : DWORD;
  cstat : TComStat;
begin
  actual := 10;
  if not ReadFile(fFile, Ptr^, Cardinal(count), actual, nil) then
  begin
    GetLastError();
    ClearCommError(fFile, Errors, @cstat);
    Result := 0;
  end
  else
  begin
    Result := actual;
  end;
end;

function TNQCSerial.SetDTR(bDTR: Boolean): Boolean;
begin
  if bDTR then
    Result := EscapeCommFunction(fFile, Windows.SETDTR)
  else
    Result := EscapeCommFunction(fFile, CLRDTR);
end;

function TNQCSerial.SetRTS(bRTS: Boolean): Boolean;
begin
  if bRTS then
    Result := EscapeCommFunction(fFile, Windows.SETRTS)
  else
    Result := EscapeCommFunction(fFile, CLRRTS);
end;

function TNQCSerial.SetSpeed(speed, opts: integer): Boolean;
var
  dcb : TDCB;
begin
  Result := False;
  FillChar(dcb, sizeof(TDCB), 0);
  dcb.DCBlength := sizeof(TDCB);
  if not GetCommState(fFile, dcb) then Exit;
  dcb.BaudRate  := speed;
  dcb.XonLim    := 100;
  dcb.XoffLim   := 100;
  dcb.ByteSize  := 8 - (opts and kPSerial_DataMask);
  dcb.Parity    := (opts and kPSerial_ParityMask) shr kPSerial_ParityShift;
  dcb.StopBits  := (opts and kPSerial_StopMask) shr kPSerial_StopShift;
  dcb.XonChar   := Chr(1);
  dcb.XoffChar  := Chr(2);
  dcb.ErrorChar := Chr(0);
  dcb.EofChar   := Chr(0);
  dcb.EvtChar   := Chr(0);
  dcb.Flags     := fBinary or fDtrControlEnable;
  Result := SetCommState(fFile, dcb);
end;

function TNQCSerial.SetTimeout(timeout_ms: Integer): Boolean;
var
  timeouts : TCommTimeouts;
begin

	timeouts.ReadIntervalTimeout := 0;
	timeouts.ReadTotalTimeoutMultiplier := 0;
	timeouts.ReadTotalTimeoutConstant := 0;
	timeouts.WriteTotalTimeoutMultiplier := 0;
	timeouts.WriteTotalTimeoutConstant := 0;

	case timeout_ms of
		kPStream_NeverTimeout : ;
		0 : timeouts.ReadIntervalTimeout := MAXDWORD;
	else
		timeouts.ReadTotalTimeoutConstant := timeout_ms;
	end;
	SetCommTimeouts(fFile, timeouts);
	Result := true;
end;

function TNQCSerial.Write(Ptr : PByte; count: Integer): LongInt;
var
  actual : DWORD;
begin
  if not WriteFile(fFile, Ptr^, Cardinal(count), actual, nil) then
    Result := -1
  else
    Result := actual;
end;

{ TNQCUsbTower }

function TNQCUsbTower.Open(Name: string): Boolean;
begin
  fTimeoutValue := 5000;
  result := inherited Open(Name);
end;

function TNQCUsbTower.Read(Ptr: PByte; count: Integer): LongInt;
begin
  Result := inherited Read(Ptr, count);
end;
{
function TNQCUsbTower.Read(Ptr: PByte; count: Integer): LongInt;
var
  len, elapsed : Integer;
  actual, timebegin, timenow : DWORD;
begin
  len := 0;
  actual := 0;
	while len < count do begin
    timebegin := GetTickCount;
    while actual = 0 do begin
      ReadFile(fFile, PChar(Ptr)[len], count - len, actual, nil);
      timenow := GetTickCount;
      elapsed := timenow - timebegin;
      if elapsed > fTimeoutValue then break;
    end;
    if count = 0 then break;
    len := len + count;
  end;
  result := len;
end;
}

function TNQCUsbTower.SetDTR(bDTR: Boolean): Boolean;
begin
  // do nothing
  result := True;
end;

function TNQCUsbTower.SetRTS(bRTS: Boolean): Boolean;
begin
  // do nothing
  result := True;
end;

function TNQCUsbTower.SetSpeed(speed, opts: integer): Boolean;
begin
  // do nothing
  result := True;
end;

function TNQCUsbTower.SetTimeout(timeout_ms: Integer): Boolean;
begin
  fTimeoutValue := timeout_ms;
  result := True;
end;

end.
