unit FakeSpirit;

interface

uses
  Classes, SysUtils, rcx_link, Windows;

const
  K_CYBER = 'CyberMaster';
  K_SCOUT = 'Scout';
  K_RCX = 'RCX';
  K_RCX2 = 'RCX2';

const
  // remote commands
  kRemoteKeysReleased = $0000;
  kRemotePBMessage1   = $0100;
  kRemotePBMessage2   = $0200;
  kRemotePBMessage3   = $0400;
  kRemoteOutAForward  = $0800;
  kRemoteOutBForward  = $1000;
  kRemoteOutCForward  = $2000;
  kRemoteOutABackward = $4000;
  kRemoteOutBBackward = $8000;
  kRemoteOutCBackward = $0001;
  kRemoteSelProgram1  = $0002;
  kRemoteSelProgram2  = $0004;
  kRemoteSelProgram3  = $0008;
  kRemoteSelProgram4  = $0010;
  kRemoteSelProgram5  = $0020;
  kRemoteStopOutOff   = $0040;
  kRemotePlayASound   = $0080;

type
  EMessageInvalid = class(Exception)
  public
    constructor Create(aMsg : string);
  end;

  TTransmitLevel = (tlNear, tlFar);
  TLSSource = (lsVariable, lsError, lsConstant);
  TThresholdValue = 0..1020;
  TBlinkTimeValue = 1..32767;
  TTimerNumber = 0..3; // rcx2 has 4, scout only has 3
  TCounterNumber = 0..2; // rcx2 has 3 counters, scout only has 2
  TTCSource = (tcVariable, tcError1, tcConstant, tcError2, tcRandom);
  TScoutMotion = (smNone, smForward, smZigZag, smCircleRight, smCircleLeft, smLoopA, smLoopB, smLoopAB);
  TScoutTouch = (stIgnore, stReverse, stAvoid, stWaitFor, stBrake);
  TScoutLight = (slIgnore, slSeekLight, slSeekDark, slAvoid, slWaitFor, slBrake);
  TScoutScale = (ssShort, ssMedium, ssLong);
  TScoutEffects = (seNone, seBug, seAlarm, seRandom, seScience);
  TSoundSetNumber = 0..5;
  TGlobalOutAction = (goaFloat, goaOff, goaOn);
  TGlobalDirAction = (gdaBackward, gdaSwitch, gdaForward);
  TMotorsNum = 1..7;

  TFakeSpirit = class
  private
    fLink : TAutoLink;
    FRCXType: TRCXType;
    FPortNum: TPortNum;
    fOnDownloadDone : TNotifyEvent;
    fOnDownloadStatus : TDownloadStatusEvent;
    fVerbose: boolean;
    fFastMode: boolean;
    fRxTimeout: Word;
    fDatalog : TStrings;
    fMemMap : TStrings;
    fMemData : TStrings;
    fOnDownloadStart: TNotifyEvent;
    fAutoClose: boolean;
    fOnOpenStateChanged: TNotifyEvent;
    procedure SetPortNum(const Value: TPortNum);
    procedure SetVerbose(const Value: boolean);
    procedure SetRCXType(const Value: TRCXType);
    procedure SetFastMode(const Value: boolean);
    function GetPortName: string;
    function GetNicePortName: string;
    function GetIsOpen: boolean;
    function GetRCXTypeName: string;
  protected
    procedure DoDownloadDone;
    procedure DoDownloadStart;
    procedure DoDownloadStatus(cur, total : Integer; var Abort : boolean);
    procedure DoOpenStateChanged;
    procedure HandleDownloadDone(Sender : TObject);
    procedure HandleDownloadStart(Sender : TObject);
    procedure HandleDownloadStatus(Sender : TObject; cur, total : Integer; var Abort : boolean);
    procedure HandleOpenStateChanged(Sender : TObject);
    function  VarHelper(aOp, aVar, aSrc : Byte; aNum : integer) : boolean;
  public
    constructor Create(aType : TRCXType = 0; aPort : TPortNum = 1);
    destructor Destroy; override;

    function  Open : boolean;
    function  Close : boolean;

    // PBrick sound commands
    function PlayTone(aFreq, aTime : integer) : boolean;
    function PlaySystemSound(aSnd : integer) : boolean;

    // PBrick output control commands
    function MotorsOn(aMotorList : Byte) : boolean; // On
    function MotorsOff(aMotorList : Byte) : boolean; // Off
    function MotorsFloat(aMotorList : Byte) : boolean; // Float
    function SetFwd(aMotorList : Byte) : boolean;
    function SetRwd(aMotorList : Byte) : boolean;
    function SwitchDirection(aMotorList : Byte) : boolean; // AlterDir
    function SetMotorPower(aMotorList : Byte; aSrc, aNum : integer) : boolean; // SetPower

    // PBrick input control commands
    function SetSensorType(aNum, aType : integer) : boolean;
    function SetSensorMode(aNum, aMode, aSlope : integer) : boolean;
    function ClearSensorValue(aNum : integer) : boolean;

    // general
    function TowerExists : boolean; // TowerAlive / TowerAndCableConnected
    function Ping : boolean;
    function PrepareBrick : boolean;
    function UnlockFirmware : boolean;
    function UnlockBrick : string;  // UnlockPBrick
    function DownloadMemoryMap : TStrings; // MemMap
    function PowerDownTime(aTime : integer) : boolean; // PBPowerDownTime
    function BatteryLevel : integer; // PBBattery
    function BrickAlive : boolean; // PBAliveOrNot
    function Shutdown : boolean; // PBTurnOff
    function Sleep(aVal : integer) : boolean;
	  function Version(var rom : Cardinal; var ram : Cardinal) : boolean;
    function TransmitPower(aLevel : TTransmitLevel) : boolean; // PBTxPower

    function Poll(aSrc, aNum : integer) : integer;
    function StartTask(aTask : integer) : boolean;
    function StopTask(aTask : integer) : boolean;
    function StopAllTasks : boolean;
    function DeleteTask(aTask : integer) : boolean;
    function DeleteAllTasks : boolean;
    function DeleteSub(aSub : integer) : boolean;
    function DeleteAllSubs : boolean;
    function ClearTimer(aNum : integer) : boolean;
    function ClearMemory : boolean;

    function GetOutputStatus(aOut : integer) : integer;
    function GetVariableValue(aVar: integer): integer;
    function GetInputValue(aIn: integer): integer;
    function GetMessageValue(aNum : integer) : integer;
    function GetTimerValue(aNum : integer) : integer;
    function GetCounterValue(aNum : integer) : integer;

    // PBrick arithmetic/logical commands
    function SetVar(aVar, aSrc, aNum : integer) : boolean;
    function SumVar(aVar, aSrc, aNum : integer) : boolean;
    function SubVar(aVar, aSrc, aNum : integer) : boolean;
    function DivVar(aVar, aSrc, aNum : integer) : boolean;
    function MulVar(aVar, aSrc, aNum : integer) : boolean;
    function SgnVar(aVar, aSrc, aNum : integer) : boolean;
    function AbsVar(aVar, aSrc, aNum : integer) : boolean;
    function AndVar(aVar, aSrc, aNum : integer) : boolean;
    function OrVar(aVar, aSrc, aNum : integer) : boolean;

    // communication to brick
    function SendRawCommand(aCmd : string; bRetry : boolean) : string;
    function SendRemote(aEvent : string; aRepeat : integer = 1) : boolean; overload;
    function SendRemote(aEvent : Word; aRepeat : integer = 1) : boolean; overload;
    function SendMessage(aMsg : integer) : boolean;

    // RCX only
    function SelectProgram(aProg : integer) : boolean; // SelectPrgm
    function SelectDisplay(aSrc, aNumber : integer) : boolean;
    function SetWatch(aHrs, aMins : integer) : boolean; overload;
    function SetWatch(aTime : string) : boolean; overload;
    function DownloadFirmware(aFile : string; bFast : boolean; bUnlock : boolean) : boolean;
    function SetDatalog(aSize : integer) : boolean;
    function DatalogNext(aSrc, aNum : integer) : boolean;
    function UploadDatalog(aFrom, aSize : integer) : TStrings; overload;
    function UploadDatalog(bVerbose : boolean) : TStrings; overload;

    // CM only methods
    function Drive(aLeft, aRight : integer) : boolean;
    function ClearTachoCounter(aMotorList : Byte) : boolean;
    function OnWait(aMotorList : Byte; aNum : integer; aTime : Byte) : boolean;
    function OnWaitDifferent(aMotorList : Byte;
      aNum0, aNum1, aNum2 : integer; aTime : Byte) : boolean;

    // Scout only methods
    function Scout(aVal : integer) : boolean; overload;
    function Scout(bPower : boolean = true) : boolean; overload;
    function CalibrateLightSensor : boolean;
    function SendVLL(aSrc, aNum : integer) : boolean;
    function SetFeedback(src, val : integer) : boolean;
    function SetLightSensorUpperThreshold(src : TLSSource; val : TThresholdValue) : boolean;
    function SetLightSensorLowerThreshold(src : TLSSource; val : TThresholdValue) : boolean;
    function SetLightSensorHysteresis(src : TLSSource; val : TThresholdValue) : boolean;
    function SetLightSensorBlinkTime(src : TLSSource; val : TBlinkTimeValue) : boolean;
    function SetTimerLimit(num : TTimerNumber; src : TTCSource; val : integer) : boolean;
    function SetCounterLimit(num : TCounterNumber; src : TTCSource; val : integer) : boolean;
    function ScoutLight(bOn : boolean) : boolean;
    function ScoutRules(motion : TScoutMotion; touch : TScoutTouch;
      light : TScoutLight; time : TScoutScale; fx : TScoutEffects) : boolean;
    function ScoutSound(bSoundEnable : boolean; bSoundOff : boolean; aNum : TSoundSetNumber) : boolean;

    // RCX2 & Scout methods
    function PollMemory(address, size : integer) : TStrings;
    function SetGlobalOutput(motors : TMotorsNum; action : TGlobalOutAction) : boolean;
    function SetGlobalDirection(motors : TMotorsNum; action : TGlobalDirAction) : boolean;
    function SetMaxPower(motors : TMotorsNum; src, num : integer) : boolean;
    function IncCounter(num : TCounterNumber) : boolean;
    function DecCounter(num : TCounterNumber) : boolean;
    function ClearCounter(num : TCounterNumber) : boolean;

    // RCX2 only methods
    function ClearSound : boolean;
    function MuteSound : boolean;
    function UnmuteSound : boolean;
    function SendUARTData(start, size : integer) : boolean;
    function SetEvent(enum, snum, etype : integer) : boolean;
    function CalibrateEvent(enum, upper, lower, hysteresis : integer) : boolean;
    function ClearAllEvents : boolean;
    function ViewSourceValue(prec, src, value : integer) : boolean;

    // properties
    property  IsOpen : boolean read GetIsOpen;
    property  FastMode : boolean read fFastMode write SetFastMode;
    property  RCXType : TRCXType read FRCXType write SetRCXType;
    property  PortNum : TPortNum read FPortNum write SetPortNum;
    property  PortName : string read GetPortName;
    property  NicePortName : string read GetNicePortName;
    property  RCXTypeName : string read GetRCXTypeName;
    property  RxTimeout : Word read fRxTimeout write fRxTimeout;
    property  VerboseMode : boolean read fVerbose write SetVerbose;
    property  AutoClose : boolean read fAutoClose write fAutoClose;
    property  DataLog : TStrings read fDataLog;
    property  MemoryMap : TStrings read fMemMap;
    property  OnDownloadStart : TNotifyEvent read fOnDownloadStart write fOnDownloadStart;
    property  OnDownloadDone : TNotifyEvent read fOnDownloadDone write fOnDownloadDone;
    property  OnDownloadStatus : TDownloadStatusEvent read FOnDownloadStatus write FOnDownloadStatus;
    property  OnOpenStateChanged : TNotifyEvent read fOnOpenStateChanged write fOnOpenStateChanged;
  end;

var
  FakeRCX : TFakeSpirit;

  function GetRCXErrorString(iErrCode : Integer) : string;

implementation

uses
  RcxLog, srecord, rcx_cmd, rcx_constants, NQCSerial;


function GetRCXErrorString(iErrCode : Integer) : string;
begin
  case iErrCode of
    kRCX_OpenSerialError : Result := 'serial port could not be opened and/or configured';
    kRCX_IREchoError     : Result := 'no echo received from IR tower';
    kRCX_ReplyError      : Result := 'no (or invalid) reply from RCX';
    kRCX_RequestError    : Result := 'attempt to send too much data';
    kRCX_FileError       : Result := 'could not open/read/write file';
    kRCX_FormatError     : Result := 'unknown file format';
    kRCX_AbortError      : Result := 'canceled by RCX_Link::DownloadProgress()';
    kRCX_MemFullError    : Result := 'not enough room in RCX program memory';
  else
    Result := '';
  end;
end;

{ TFakeSpirit }

constructor TFakeSpirit.Create(aType : TRCXType; aPort: TPortNum);
begin
  inherited Create;
  fFastMode  := False;
  fAutoClose := False;
  FRCXType   := aType;
  FPortNum   := aPort;
  fRxTimeout := 400;

  fLink          := TAutoLink.Create;
  fLink.FastMode := fFastMode;
  fLink.PortNum  := FPortNum;
  fLink.Target   := FRCXType;

  fLink.OnDownloadStart    := HandleDownloadStart;
  fLink.OnDownloadDone     := HandleDownloadDone;
  fLink.OnDownloadStatus   := HandleDownloadStatus;
  fLink.OnOpenStateChanged := HandleOpenStateChanged;

  fDatalog := TStringList.Create;
  fMemMap  := TStringList.Create;
  fMemData := TStringList.Create;
end;

destructor TFakeSpirit.Destroy;
begin
  Close;
  fLink.Free;
  fDatalog.Free;
  fMemMap.Free;
  fMemData.Free;
  inherited Destroy;
end;


procedure TFakeSpirit.DoDownloadDone;
begin
  if Assigned(fOnDownloadDone) then
    fOnDownloadDone(self);
end;

function TFakeSpirit.DownloadMemoryMap: TStrings;
var
  bOpen : boolean;
begin
  fMemMap.Clear;
  bopen := fLink.Open;
  try
    if bopen then
      fLink.GetMemMap(fMemMap);
  finally
    if fAutoClose then fLink.Close;
  end;
  result := fMemMap;
end;

function TFakeSpirit.PowerDownTime(aTime: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_AutoOffOp, aTime)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SelectDisplay(aSrc, aNumber: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeDisplay(aSrc, aNumber)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SelectProgram(aProg: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SelectProgramOp, Byte(aProg-1)),true) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetWatch(aHrs, aMins: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SetWatchOp, Byte(aHrs), Byte(aMins))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetWatch(aTime: string): boolean;
var
  H, M, S, MS : Word;
  t : integer;
begin
  if LowerCase(aTime) = 'now' then
  begin
    DecodeTime(Now, H, M, S, MS);
  end
  else
  begin
    // string is assumed to be HHMM
    t := StrToIntDef(aTime, 0);
    H := t div 100;
    M := t mod 100;
  end;
  result := SetWatch(H, M);
end;

function TFakeSpirit.UnlockFirmware: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    // sending a simple command will unlock the firmware if necessary
    result := fLink.Send(cmd.MakePing) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.UnlockBrick: string;
var
  cmd : TRcxCmd;
  cnt, i : integer;
begin
  cmd := TRcxCmd.Create;
  try
    cnt := fLink.Send(cmd.SetVal(kRCX_UnlockPBrick, 1, 3, 5, 7, $b));
    if cnt >= kRCX_Ok then
    begin
      for i := 0 to cnt - 1 do
        result := result + IntToStr(fLink.GetReplyByte(i));
    end;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.BatteryLevel: integer;
var
  bopen : boolean;
begin
  result := -1;
  bopen := fLink.Open;
  try
    if bopen then
      result := fLink.GetBatteryLevel;
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.BrickAlive: boolean;
begin
  result := true;
  if RCXType = rtScout then
  begin
    // the scout doesn't respond to pinging unless it is unlocked
    result := UnlockFirmware;
  end;
  if result then
    result := Ping;
end;

function TFakeSpirit.Shutdown: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_PBTurnOffOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.TransmitPower(aLevel: TTransmitLevel): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_IRModeOp, Ord(aLevel))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.Poll(aSrc, aNum: integer): integer;
var
  cmd : TRcxCmd;
  val, cnt : integer;
begin
  cmd := TRcxCmd.Create;
  try
    val := integer((aSrc shl 16) or (aNum and $FFFF));
    cnt := fLink.Send(cmd.MakeRead(val));
    if cnt <> 2 then
    begin
      result := -1;
      Exit;
    end;
    result := Cardinal(fLink.GetReplyByte(1) shl 8) + fLink.GetReplyByte(0);
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.StartTask(aTask: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_StartTaskOp, aTask)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.StopTask(aTask: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeStopTask(aTask)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DeleteAllTasks: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DeleteTasksOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DeleteTask(aTask: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DeleteTaskOp, aTask)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.StopAllTasks: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_StopAllOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DeleteSub(aSub: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DeleteSubOp, aSub)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DeleteAllSubs: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DeleteSubsOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.MotorsOn(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputMode(aMotorList, kRCX_OutputOn)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.MotorsOff(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputMode(aMotorList, kRCX_OutputOff)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.MotorsFloat(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputMode(aMotorList, kRCX_OutputFloat)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetFwd(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputDir(aMotorList, kRCX_OutputForward)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetRwd(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputDir(aMotorList, kRCX_OutputBackward)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SwitchDirection(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOutputDir(aMotorList, kRCX_OutputToggle)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetMotorPower(aMotorList: Byte; aSrc, aNum: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_OutputPowerOp, aMotorList, aSrc, aNum)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.Drive(aLeft, aRight: integer): boolean;
var
  cmd : TRcxCmd;
begin
  // CM only
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeDrive(aLeft, aRight)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearTachoCounter(aMotorList: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  // CM only
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearTachoOp, aMotorList)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.PlayTone(aFreq, aTime: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakePlayTone(aFreq, aTime)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.PlaySystemSound(aSnd: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakePlaySound(aSnd)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetSensorType(aNum, aType: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeInputType(aNum, aType)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetSensorMode(aNum, aMode, aSlope: integer): boolean;
var
  cmd : TRcxCmd;
  val : Byte;
begin
  cmd := TRcxCmd.Create;
  try
    val := ((aMode and $7) shl 5) or (aSlope and $F);
    result := fLink.Send(cmd.MakeInputMode(aNum, val)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearSensorValue(aNum: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearSensorOp, aNum)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearTimer(aNum: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearTimerOp, aNum)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.VarHelper(aOp, aVar, aSrc : Byte; aNum : Integer): boolean;
var
  cmd : TRcxCmd;
  l, h : Byte;
begin
  cmd := TRcxCmd.Create;
  try
    l := aNum and $FF;
    h := (aNum shr 8) and $FF;
    result := fLink.Send(cmd.SetVal(aOp, aVar, aSrc, l, h)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_SetVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.SumVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_SumVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.SubVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_SubVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.DivVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_DivVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.MulVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_MulVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.SgnVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_SgnVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.AbsVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_AbsVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.AndVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_AndVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.OrVar(aVar, aSrc, aNum: integer): boolean;
begin
  result := VarHelper(kRCX_OrVarOp, aVar, aSrc, aNum);
end;

function TFakeSpirit.DatalogNext(aSrc, aNum: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DatalogOp, aSrc, aNum)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetDatalog(aSize: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeSetDatalog(aSize)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.UploadDatalog(aFrom, aSize: integer): TStrings;
var
  cmd : TRcxCmd;
  response, i : integer;
  buf : array of Byte;
begin
  result := fDataLog;
  fDataLog.Clear;
  cmd := TRcxCmd.Create;
  try
    response := fLink.Send(cmd.MakeUploadDatalog(aFrom, aSize));
    if response > 0 then
    begin
      System.SetLength(buf, response);
      fLink.GetReply(@buf, response);
      for i := 0 to response - 1 do
      begin
        fDataLog.Add(Format('%d', [buf[i]]));
      end;
    end;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.UploadDatalog(bVerbose: boolean): TStrings;
var
  log : TRcxLog;
  i : integer;
  response : integer;
begin
  result := fDataLog;
  fDataLog.Clear;
  log := TRcxLog.Create;
  try
    if fLink.Open then
    begin
      response := log.Upload(fLink);
      if response < 0 then Exit;
      for i := 0 to log.Length - 1 do
      begin
        fDataLog.Add(log.SPrintEntry(i, bVerbose));
      end;
    end;
  finally
    if fAutoClose then fLink.Close;
    log.Free;
  end;
end;

procedure TFakeSpirit.DoDownloadStatus(cur, total : Integer; var Abort : boolean);
begin
  Abort := False;
  if Assigned(FOnDownloadStatus) then
    FOnDownloadStatus(Self, cur, total, Abort);
end;

function TFakeSpirit.PrepareBrick: boolean;
begin
  result := UnlockFirmware;
{
  if RCXType = rtScout then
  begin
    result := Scout(true);
  end;
}
end;

function TFakeSpirit.Scout(bPower : boolean = true) : boolean;
var
  cmd : TRcxCmd;
  val : Byte;
begin
  cmd := TRcxCmd.Create;
  try
    if bPower then
      val := $80
    else
      val := $00;
    result := fLink.Send(cmd.SetVal(kRCX_ScoutOp, val)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.Close: boolean;
begin
  if IsOpen then
  begin
    result := fLink.Close;
  end
  else
    result := true;
end;

function TFakeSpirit.Open: boolean;
begin
  result := IsOpen;
  if not result then begin
    result := fLink.Open;
  end;
end;

function TFakeSpirit.ClearMemory: boolean;
var
  cmd : TRcxCmd;
  response, i : integer;
begin
  result := False;
  cmd := TRcxCmd.Create;
  try
    response := fLink.Send(cmd.SetVal(kRCX_StopAllOp));
    if response < 0 then Exit;

    if (fRCXType = rtRCX) or (fRCXType = rtRCX2) then
    begin
      for i := 1 to 4 do
      begin
        response := fLink.Send(cmd.SetVal(kRCX_SelectProgramOp, i));
        if response < 0 then Exit;

        response := fLink.Send(cmd.MakeDeleteSubs);
        if response < 0 then Exit;

        response := fLink.Send(cmd.MakeDeleteTasks);
        if response < 0 then Exit;
      end;
      response := fLink.Send(cmd.SetVal(kRCX_SetDatalogOp, 0, 0));
    end
    else
    begin
      response := fLink.Send(cmd.MakeDeleteSubs);
      if response < 0 then Exit;

      response := fLink.Send(cmd.MakeDeleteTasks);
      if response < 0 then Exit;
    end;

    result := response >= 0;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.Sleep(aVal: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_AutoOffOp, Byte(aVal))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SendRawCommand(aCmd: string; bRetry: boolean): string;
var
  len, response, i : integer;
  cmd : TRcxCmd;
  bStr : string;
  tmpByte : Byte;
  orig : PByte;
begin
  result := '';
  len := Length(aCmd);
  if (len mod 2) = 0 then
  begin
    len := len div 2;
    cmd := TRcxCmd.Create;
    try
      cmd.SetLength(len);
      orig := cmd.GetBody;
      while aCmd <> '' do
      begin
        bStr := Copy(aCmd, 1, 2);
        Delete(aCmd, 1, 2);
        tmpByte := StrToIntDef('$' + bStr, 0);
        orig^ := tmpByte;
        inc(orig);
      end;
      response := fLink.Send(cmd, bRetry);
      if response > 0 then
      begin
        for i := 0 to response - 1 do
        begin
          result := result + Format('%02x ', [fLink.GetReplyByte(i)]);
        end;
      end;
    finally
      cmd.Free;
      if fAutoClose then fLink.Close;
    end;
  end;
end;

function TFakeSpirit.SendRemote(aEvent: string; aRepeat: integer): boolean;
var
  val : integer;
begin
  result := False;
  if aRepeat <= 0 then Exit;
  if Length(aEvent) <> 4 then Exit;
  val := StrToIntDef('$' + Copy(aEvent, 1, 4), 0);
  result := SendRemote(val, aRepeat);
end;

function TFakeSpirit.SendRemote(aEvent: Word; aRepeat: integer): boolean;
var
  cmd : TRcxCmd;
begin
  result := False;
  if aRepeat <= 0 then Exit;
  cmd := TRcxCmd.Create;
  try
    // open, sync, send
    result := fLink.Open;
    try
      if not result then Exit;

      result := fLink.Sync(false) >= kRCX_OK;
      if not result then Exit;

      cmd.SetVal(kRCX_Remote, Lo(aEvent), Hi(aEvent));
      while aRepeat > 0 do
      begin
        fLink.Send1(cmd.GetBody, cmd.GetLength);
        dec(aRepeat);
      end;
    finally
      if fAutoClose then fLink.Close;
    end;
    result := true;
  finally
    cmd.Free;
  end;
end;

function TFakeSpirit.SendMessage(aMsg: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    cmd.SetVal(kRCX_Message, Byte(aMsg));
    result := fLink.Send1(cmd.GetBody, cmd.GetLength) >= kRCX_OK;
//    result := fLink.Send(cmd.SetVal(kRCX_Message, Byte(aMsg)), false) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetCounterValue(aNum: integer): integer;
begin
  try
    result := fLink.GetCounterValue(aNum);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetTimerValue(aNum: integer): integer;
begin
  try
    result := fLink.GetTimerValue(aNum);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetMessageValue(aNum: integer): integer;
begin
  try
    result := fLink.GetMessageValue(aNum);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetInputValue(aIn : integer) : integer;
begin
  try
    result := fLink.GetInputValue(aIn);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetVariableValue(aVar : integer) : integer;
begin
  try
    result := fLink.GetVariableValue(aVar);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.GetOutputStatus(aOut: integer): integer;
begin
  try
    result := fLink.GetOutputStatus(aOut);
  finally
    if fAutoClose then fLink.Close;
  end;
end;

procedure TFakeSpirit.SetPortNum(const Value: TPortNum);
begin
  FPortNum := Value;
  fLink.PortNum := Value;
end;

procedure TFakeSpirit.SetVerbose(const Value: boolean);
begin
  fVerbose := Value;
  fLink.Verbose := Value;
end;

procedure TFakeSpirit.SetRCXType(const Value: TRCXType);
begin
  FRCXType := Value;
  fLink.Target := Value;
end;

function TFakeSpirit.Scout(aVal: integer): boolean;
var
  bPower : boolean;
begin
  bPower := aVal <> 0;
  result := Scout(bPower);
end;

function TFakeSpirit.Version(var rom, ram: Cardinal): boolean;
var
  bopen : boolean;
begin
  result := false;
  bopen := fLink.Open;
  try
    if bopen then
      result := fLink.GetVersion(rom, ram) >= kRCX_OK;
  finally
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DownloadFirmware(aFile: string; bFast : Boolean; bUnlock : boolean): boolean;
const
  kMaxFirmware = 65536;
var
  F : TextFile;
  srec : TSRecord;
  ok : boolean;
  response : integer;
  rom, ram : Cardinal;
begin
  result := False;
  try
    AssignFile(F, aFile);
    srec := TSRecord.Create;
    try
      Reset(F);
      try
        ok := srec.Read(F, kMaxFirmware);
        if not ok then
        begin
          result := False;
          Exit;
        end;
      finally
        CloseFile(F);
      end;
	    result := fLink.Open;
      try
        if not result then Exit;
        response := fLink.SendFirmware(srec.GetData, srec.GetLength, srec.GetStart, bFast);
        if response < 0 then
        begin
          result := False;
          Exit;
        end;
        // this unlocks the firmware
        response := fLink.GetVersion(rom, ram);
        result := response >= 0;
      finally
        if fAutoClose then fLink.Close;
      end;
    finally
      srec.Free;
    end;
  except
  end;
end;

procedure TFakeSpirit.DoDownloadStart;
begin
  if Assigned(fOnDownloadStart) then
    fOnDownloadStart(self);
end;

procedure TFakeSpirit.HandleDownloadDone(Sender: TObject);
begin
  DoDownloadDone;
end;

procedure TFakeSpirit.HandleDownloadStart(Sender: TObject);
begin
  DoDownloadStart;
end;

procedure TFakeSpirit.HandleDownloadStatus(Sender: TObject;
  cur, total : Integer; var Abort: boolean);
begin
  DoDownloadStatus(cur, total, Abort);
end;

function TFakeSpirit.Ping: boolean;
var
  cmd : TRcxCmd;
{
  ot : integer;
  odt : boolean;
}
begin
  cmd := TRcxCmd.Create;
  try
{
    ot  := fLink.RxTimeout;
    odt := fLink.DynamicTimeout;
    try
      fLink.RxTimeout := Self.RxTimeout;
}
      result := fLink.Send(cmd.MakePing) >= kRCX_OK;
{
    finally
      fLink.RxTimeout := ot;
      fLink.DynamicTimeout := odt;
    end;
}
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.TowerExists: boolean;
var
  cmd : TRcxCmd;
  val : integer;
begin
  cmd := TRcxCmd.Create;
  try
    val := fLink.Send(cmd.MakePing, False);
    result := (val <> kRCX_OpenSerialError) and (val <> kRCX_IREchoError);
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.OnWait(aMotorList: Byte; aNum: integer;
  aTime: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  // CM only
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.MakeOnWait(aMotorList, aNum, aTime)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.OnWaitDifferent(aMotorList: Byte; aNum0, aNum1,
  aNum2: integer; aTime: Byte): boolean;
var
  cmd : TRcxCmd;
begin
  // CM only
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(
      cmd.MakeOnWaitDifferent(aMotorList, aNum0, aNum1, aNum2, aTime)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.CalibrateLightSensor: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_LSCalibrateOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SendVLL(aSrc, aNum: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_VLLOp, aSrc, aNum)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetFeedback(src, val: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SetFeedbackOp, src, Lo(val), Hi(val))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetLightSensorUpperThreshold(src : TLSSource; val: TThresholdValue): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_LSUpperThreshOp, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetLightSensorLowerThreshold(src : TLSSource; val: TThresholdValue): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_LSLowerThreshOp, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetLightSensorHysteresis(src: TLSSource; val: TThresholdValue): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_LSHysteresisOp, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetLightSensorBlinkTime(src: TLSSource; val: TBlinkTimeValue): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_LSBlinkTimeOp, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
   finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetTimerLimit(num: TTimerNumber; src: TTCSource; val: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SetTimerLimitOp, num, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetCounterLimit(num: TCounterNumber; src: TTCSource; val: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SetCounterOp, num, Ord(src), Lo(Ord(val)), Hi(Ord(val)))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ScoutLight(bOn: boolean): boolean;
var
  cmd : TRcxCmd;
  val : integer;
begin
  cmd := TRcxCmd.Create;
  try
    val := $0;
    if bOn then
      val := $80;
    result := fLink.Send(cmd.SetVal(kRCX_LightOp, val)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ScoutRules(motion: TScoutMotion; touch: TScoutTouch;
  light: TScoutLight; time: TScoutScale; fx: TScoutEffects): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ScoutRulesOp, Ord(motion),
      Ord(touch), Ord(light), Ord(time), Ord(fx))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ScoutSound(bSoundEnable, bSoundOff: boolean;
  aNum: TSoundSetNumber): boolean;
var
  cmd : TRcxCmd;
  num : integer;
begin
  cmd := TRcxCmd.Create;
  try
    num := aNum;
    if bSoundEnable then
      inc(num, $80);
    if bSoundOff then
      inc(num, $40);
    result := fLink.Send(cmd.SetVal(kRCX_SoundOp, num)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.PollMemory(address, size: integer): TStrings;
var
  cmd : TRcxCmd;
  cnt, i : integer;
  buf : array of Byte;
begin
  result := fMemData;
  fMemData.Clear;
  cmd := TRcxCmd.Create;
  try
    cnt := fLink.Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(address), Hi(address), size));
    if cnt > 0 then
    begin
      System.SetLength(buf, cnt);
      fLink.GetReply(@buf, cnt);
      for i := 0 to cnt - 1 do
      begin
        fMemData.Add(Format('%d', [buf[i]]));
      end;
    end;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetGlobalOutput(motors: TMotorsNum; action: TGlobalOutAction): boolean;
var
  cmd : TRcxCmd;
  act : integer;
begin
  case action of
    goaOff : act := $40;
    goaOn  : act := $80;
  else
    act := 0;
  end;
  inc(act, motors);
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_GOutputModeOp, act)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetGlobalDirection(motors: TMotorsNum; action: TGlobalDirAction): boolean;
var
  cmd : TRcxCmd;
  act : integer;
begin
  case action of
    gdaSwitch  : act := $40;
    gdaForward : act := $80;
  else
    act := 0;
  end;
  inc(act, motors);
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_GOutputDirOp, act)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetMaxPower(motors : TMotorsNum; src, num: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_GOutputPowerOp, motors, Ord(src), Ord(num) )) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.IncCounter(num: TCounterNumber): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_IncCounterOp, num)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.DecCounter(num: TCounterNumber): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_DecCounterOp, num)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearCounter(num: TCounterNumber): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearCounterOp, num)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearSound: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearSound)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.MuteSound: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    if RCXType = rtScout then
    begin
      result := ScoutSound(True, True, 0);
    end
    else
      result := fLink.Send(cmd.SetVal(kRCX_MuteSoundOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.UnmuteSound: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    if RCXType = rtScout then
    begin
      result := ScoutSound(True, False, 0);
    end
    else
      result := fLink.Send(cmd.SetVal(kRCX_UnmuteSoundOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SendUARTData(start, size: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    cmd.SetVal(kRCX_SendUARTDataOp, start, size);
    result := fLink.Send1(cmd.GetBody, cmd.GetLength) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.SetEvent(enum, snum, etype: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_SetEventOp, enum, snum, etype)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.CalibrateEvent(enum, upper, lower, hysteresis: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_CalibrateEventOp, enum, upper, lower, hysteresis)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ClearAllEvents: boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ClearAllEventsOp)) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

function TFakeSpirit.ViewSourceValue(prec, src, value: integer): boolean;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := fLink.Send(cmd.SetVal(kRCX_ViewSourceValOp, 0, prec, src, Lo(value), Hi(value))) >= kRCX_OK;
  finally
    cmd.Free;
    if fAutoClose then fLink.Close;
  end;
end;

procedure TFakeSpirit.SetFastMode(const Value: boolean);
begin
  if fFastMode <> Value then
  begin
    Close;
    fFastMode := Value;
    fLink.FastMode := Value;
  end;
end;

function TFakeSpirit.GetPortName: string;
begin
  Result := fLink.PortName;
end;

function TFakeSpirit.GetNicePortName: string;
begin
  Result := Copy(PortName, 5, 15);
end;

function TFakeSpirit.GetIsOpen: boolean;
begin
  result := fLink.IsOpen;
end;

procedure TFakeSpirit.DoOpenStateChanged;
begin
  if Assigned(FOnOpenStateChanged) then
    FOnOpenStateChanged(self);
end;

procedure TFakeSpirit.HandleOpenStateChanged(Sender: TObject);
begin
  DoOpenStateChanged;
end;

function TFakeSpirit.GetRCXTypeName: string;
const
  BrickNames : array[0..3] of String = (K_RCX, K_CYBER, K_SCOUT, K_RCX2);
begin
  Result := BrickNames[RCXType];
end;

{ EMessageInvalid }

resourcestring
  sErrorInvalidMessage = 'Invalid raw message';

constructor EMessageInvalid.Create(aMsg: string);
begin
  inherited Create(sErrorInvalidMessage + ' ' + aMsg);
end;

initialization
  FakeRCX := TFakeSpirit.Create;

finalization
  FakeRCX.Free;

end.
