unit rcx_link;

interface

uses
  Classes, SysUtils, NQCSerial, Windows, rcx_cmd;

const
  rtRCX = 0;
  rtCybermaster = 1;
  rtScout = 2;
  rtRCX2 = 3;

type
  TDownloadStatusEvent = procedure(Sender : TObject; cur, total : Integer; var Abort : boolean) of object;
  TRCXType = 0..3;
  TPortNum = 1..12;

  TRcxLink = class
  private
    fSerial : TNQCSerial;
    fTarget : TRCXType;
    fFastMode: boolean;
    fTxData : PByte;
    fTxLength : integer;
    fTxLastCommand : Byte;

    fRxData : PByte;
    fRxLength : integer;
    fRxExpected : integer;
    fRxState : integer;

    fResult : integer;

    fVerbose : boolean;
    fPredictive : boolean;
    fSynced : boolean;
    fRxTimeout : integer;
    fDynamicTimeout : boolean;

    fDownloadTotal : integer;
    fDownloadSoFar : integer;
    FOnDownloadStatus: TDownloadStatusEvent;
    fOnDownloadDone: TNotifyEvent;
    fOnDownloadStart: TNotifyEvent;
    fIgnoreReply: boolean;
    fOnOpenStateChanged: TNotifyEvent;
    function  _SendFirmware(data : PByte; length, start : integer; progress : boolean) : integer;
    function  _SendFromTxBuffer(timeout : integer) : integer;
    function  FindReply(var offset : integer) : integer;
    function  IncrementProgress(delta : integer) : boolean;
    function  ValidateRxData : integer;
    procedure BuildTxData(data : PByte; length : integer; duplicateReduction : boolean);
    procedure BeginProgress(total : integer);
    procedure AdjustReply(offset, length : integer);
    procedure SetRxState(state : integer);
    procedure ProcessRxByte(b : Byte);
    procedure SetIgnoreReply(const Value: boolean);
    procedure SetFastMode(const Value: boolean);
  protected
    fLog : string;
    fPortNum: TPortNum;
    function  GetPortNum: TPortNum; virtual;
    procedure SetPortNum(const Value: TPortNum); virtual;
    procedure DumpData(ptr : PByte; length : integer);
    procedure WriteToLog(s : string);
    procedure ClearLog;
    procedure SetVerbose(b : boolean);
    function  GetVerbose : boolean;
    procedure SetTarget(target : TRCXType);
    procedure DoDownloadDone;
    procedure DoDownloadStart;
    procedure DoDownloadStatus(cur, total : Integer; var Abort : boolean);
    procedure DoOpenStateChanged;
    function  PredictReplyLength(data : PByte; len : integer) : Integer;
    procedure SetRxTimeout(t : integer);
    function  GetNQCSerial : TNQCSerial;
    property  NQCSerial : TNQCSerial read GetNQCSerial;
    function  GetPortName : string;
  public
    constructor Create;
    destructor  Destroy; override;
    function  Sync(bScoutPower : boolean = true) : integer;

    function  Send(cmd : TRcxCmd; retry : boolean = true; timeout : integer = 0) : integer; overload;
    function  Send(data : PByte; length : integer; retry : boolean = true; timeout : integer = 0) : integer; overload;
    function  Send1(data : PByte; length : integer) : integer;
    function  Receive(echo : boolean) : integer;
    function  GetReply(data : PByte; maxLength : integer) : integer;
    function  GetReplyByte(index : integer) : Byte;

    function  DownloadProgress(soFar, total : integer) : boolean; virtual;
    function  SendFragment(ftype : TRcxFragment; taskNumber : Byte; const data : PByte; length : integer; total : integer = 0) : integer;
    function  SendFirmware(const data : PByte; length, start : integer; fast : boolean) : integer;
    function  Download(const data : PByte; length, chunk : integer) : integer;
    function  GetVersion(var rom : Cardinal; var ram : Cardinal) : integer;
    function  GetMemMap(aMemMap : TStrings) : integer;
    function  GetBatteryLevel : integer;
    function  GetOutputStatus(aOut : Byte) : Integer;
    function  GetVariableValue(aVar : Byte) : Integer;
    function  GetInputValue(aIn: Byte): Integer;
    function  GetMessageValue(aNum : Byte) : Integer;
    function  GetTimerValue(aNum : Byte) : Integer;
    function  GetCounterValue(aNum : Byte) : Integer;
    function  Poll(aSrc, aNum : integer) : integer;
    function  GetValue(value : integer) : integer;
    function  Open(aPortNum : TPortNum) : boolean; overload;
    function  Close : boolean; virtual;

    property  Log : string read fLog;
    property  PortNum : TPortNum read GetPortNum write SetPortNum;
    property  PortName : string read GetPortName;
    property  Verbose : boolean read GetVerbose write SetVerbose;
    property  Target : TRcxType read fTarget write SetTarget;
    property  FastMode : boolean read fFastMode write SetFastMode;
    property  IgnoreReply : boolean read fIgnoreReply write SetIgnoreReply;
    property  RxTimeout : integer read fRxTimeout write SetRxTimeout;
    property  DynamicTimeout : boolean read fDynamicTimeout write fDynamicTimeout;
    property  OnDownloadStart : TNotifyEvent read fOnDownloadStart write fOnDownloadStart;
    property  OnDownloadDone : TNotifyEvent read fOnDownloadDone write fOnDownloadDone;
    property  OnDownloadStatus : TDownloadStatusEvent read FOnDownloadStatus write FOnDownloadStatus;
    property  OnOpenStateChanged : TNotifyEvent read fOnOpenStateChanged write fOnOpenStateChanged;
  end;

  TAutoLink = class(TRcxLink)
  private
    fOpen : boolean;
  protected
    procedure SetPortNum(const Value: TPortNum); override;
  public
    constructor Create;
    function Open : boolean; overload;
    function  Close : boolean; override;
    function Send(cmd : TRcxCmd; bRetry : boolean = true) : integer; overload;
    function  DownloadProgress(soFar, total : integer) : boolean; override;
    property IsOpen : boolean read fOpen;
  end;

function PortIsUSB(i : integer) : Boolean;

implementation

uses
  rcx_constants, scout_def;

const
  MAX_COMPORT = 8;

  kReplyState = 0;
  kIdleState  = 1;
  kSync1State = 2;
  kSync2State = 3;
  kDataState  = 4;

  // global constants
  kMaxCmdLength   = 1000;
  kMaxReplyLength	= 1000;
  kMinTimeout     = 50;
  kMaxTimeout     = 400;

  kMaxTxData         = 2 * kMaxCmdLength + 6;
  kMaxRxData         = kMaxTxData + (2 * kMaxReplyLength) + 5;
  kFragmentChunk     = 20;
  kFirmwareChunk     = 200;
  kDefaultRetryCount = 5;
  kFirmwareDelay     = 250; // milliseconds
  kNubStart          = $8000;

  kRCX_IllegalVar = -1;

  rcxnub : array[0..87] of Byte = (
    $79,$06,$00,$0f,$6b,$86,$ee,$80,$79,$06,$ee,$64,$6d,$f6,$79,$06,
    $ee,$74,$5e,$00,$3b,$9a,$0b,$87,$79,$06,$ee,$5e,$5e,$00,$06,$88,
    $7f,$d8,$72,$50,$fe,$67,$3e,$d9,$18,$ee,$6a,$8e,$ef,$51,$fe,$02,
    $6a,$8e,$ef,$06,$fe,$0d,$6a,$8e,$ee,$5e,$54,$70,$44,$6f,$20,$79,
    $6f,$75,$20,$62,$79,$74,$65,$2c,$20,$77,$68,$65,$6e,$20,$49,$20,
    $6b,$6e,$6f,$63,$6b,$3f,$00,$00);

const
  USB_TOWER = '\\.\LEGOTOWER';
  COM_PORT  = '\\.\COM';

function PortIsUSB(i : integer) : Boolean;
begin
  Result := i in [MAX_COMPORT+1..High(TPortNum)];
end;

function RCX_ERROR(res : integer) : boolean;
begin
  Result := res < 0;
end;

function FindRCXSync(data : PByte; length : integer) : integer;
var
  _end : PByte;
  ptr : PByte;
  p1, p2 : PByte;
begin
  _end := data;
  inc(_end, length - 2);

  ptr := data;
  while Integer(ptr) < Integer(_end) do
  begin
    p1 := ptr;
    inc(p1);
    p2 := p1;
    inc(p2);
    if (ptr^ = $55) and (p1^ = $FF) and (p2^ = $00) then
    begin
      result := Integer(ptr) - Integer(data) + 3;
      Exit;
    end;
    if (ptr^ = $FF) and (p1^ = $00) then
    begin
      result := Integer(ptr) - Integer(data) + 2;
      Exit;
    end;
    inc(ptr);
  end;
  result := 0;
end;

function FindCMSync(data : PByte; length : integer) : integer;
var
  _end : PByte;
  ptr : PByte;
begin
  _end := data;
  inc(_end, length);

  ptr := data;
  while Integer(ptr) < Integer(_end) do
  begin
    if ptr^ = $ff then
    begin
      result := Integer(ptr) - Integer(data) + 1;
      Exit;
    end;
    inc(ptr);
  end;
  result := 0;
end;

function VerifyReply(data : PByte; length : integer; cmd : Byte) : integer;
var
  checksum : Byte;
  ptr, _end, match, tmpD : PByte;
begin
  // always need a cmd and a checksum
  if length < 4 then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  tmpD := data;
  inc(tmpD);
  // check the cmd
  if (data^ <> ((not cmd) and $ff)) or (tmpD^ <> cmd) then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  ptr := data;
  inc(ptr, 2);
  _end := data;
  inc(_end, length - 1);
  checksum := data^;
  match := nil;

  while Integer(ptr) < Integer(_end) do
  begin
    tmpD := ptr;
    inc(tmpD);
    if ptr^ <> ((not tmpD^) and $ff) then
      break;

    if ptr^ = checksum then
      match := ptr;

    checksum := checksum + ptr^;
    inc(ptr,2);
  end;

  if match = nil then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  result := ((Integer(match) - Integer(data)) div 2) - 1;
end;

function VerifyFastReply(data : PByte; length : integer; cmd : Byte) : integer;
var
  checksum : Byte;
  ptr, _end, match : PByte;
begin
  // always need a cmd and a checksum
  if length < 2 then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  // check the cmd
  if data^ <> ((not cmd) and $ff) then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  ptr := data;
  inc(ptr);
  _end := data;
  inc(_end, length);
  checksum := data^;
  match := nil;

  while Integer(ptr) < Integer(_end) do
  begin
    if ptr^ = checksum then
      match := ptr;

    checksum := checksum + ptr^;
    inc(ptr);
  end;

  if match = nil then
  begin
    result := kRCX_ReplyError;
    Exit;
  end;

  result := ((Integer(match) - Integer(data))) - 1;
end;

{ TRcxLink }

function TRcxLink.PredictReplyLength(data : PByte; len : integer) : integer;
var
  b : Byte;
begin
  b := data^ and $F7;
  case b of
    $25, $35, $45, $75 :
      result := 1;
    $12, $30 :
      result := 2;
    $15 : result := 8;
    $a5 : result := 25;
    $20 : begin // memmap
      if fTarget = rtCybermaster then
        result := 20
      else if fTarget = rtScout then
        result := 0 // unsupported op-code
      else
        result := 188;
    end;
    $63 : begin // pollm
      if len <> 4 then
        result := 0
      else
        result := PByte(Integer(data)+3)^;
    end;
    $a4 : begin // upload datalog
      if len <> 5 then
        result := 0
      else
        result := ( PByte(Integer(data)+3)^ +
                   (Integer(PByte(Integer(data)+4)^) shl 8)) * 3;
    end;
  else
    result := 0;
  end;
end;

function TRcxLink._SendFirmware(data: PByte; length, start: integer;
  progress: boolean): integer;
var
  cmd : TRcxCmd;
  check, i : integer;
  tmpD : PByte;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if (RCX_ERROR(result)) then Exit;

    result := Send(cmd.MakeUnlock());
    if (RCX_ERROR(result)) then Exit;

    result := Send(cmd.SetVal(kRCX_BootModeOp, 1, 3, 5, 7, $b));
    if (RCX_ERROR(result)) then Exit;

    check := 0;
    tmpD  := data;
    for i := 0 to length - 1 do
    begin
      if i >= $4c00 then break;
      check := check + tmpD^;
      inc(tmpD);
    end;
    result := Send(cmd.SetVal(kRCX_BeginFirmwareOp, Byte(start), Byte(start shr 8),
      Byte(check), Byte(check shr 8), 0));
    if (RCX_ERROR(result)) then Exit;

    if progress then
      BeginProgress(length)
    else
      BeginProgress(0);
    result := Download(data, length, kFirmwareChunk);
    if (RCX_ERROR(result)) then Exit;

    // last packet is no-retry with an extra long delay
    // this gives the RCX time to respond and makes sure response doesn't get trampled
    result := Send(cmd.MakeBoot(), false, kFirmwareDelay);

    if fFastMode then
    begin
      // for some reason the response doesn't checksum in fast mode
      // just assume it was ok
      result := kRCX_OK;
    end;
  finally
    cmd.Free;
  end;
end;

function TRcxLink._SendFromTxBuffer(timeout: integer): integer;
var
  offset : integer;
  tmpRxD : PByte;
begin
  // drain serial rx buffer
  NQCSerial.SetTimeout(0);
  while NQCSerial.Read(fRxData, kMaxRxData) > 0 do
  begin
  end;

  // send command
  NQCSerial.Write(fTxData, fTxLength);
  NQCSerial.FlushWrite();
  if fVerbose then
  begin
    WriteToLog('Tx: ');
    DumpData(fTxData, fTxLength);
  end;

  if fIgnoreReply then
  begin
    result := kRCX_Ok;
    Exit;
  end;

  // get the reply
  NQCSerial.SetTimeout(timeout);
  fRxState := kReplyState;
  fRxLength := 0;

//  fRxLength := NQCSerial.Read(fRxData, kMaxRxData);
  while fRxLength < kMaxRxData do
  begin
    tmpRxD := fRxData;
    inc(tmpRxD,fRxLength);
    if NQCSerial.Read(tmpRxD, 1) <> 1 then
      break;

    inc(fRxLength);

    // check for predictive replies...doesn't seem to work with firmware download
    if fPredictive then
    begin
      result := FindReply(offset);
      if result = fRxExpected then
        break;
    end;
  end;

  if fVerbose then
  begin
    WriteToLog('Rx: ');
    DumpData(fRxData, fRxLength);
  end;

  if fRxLength = 0 then
  begin
    if fTarget = rtCybermaster then
      result := kRCX_ReplyError
    else
      result := kRCX_IREchoError;
    Exit;
  end;

  // find reply
  result := FindReply(offset);

  if not RCX_ERROR(result) then
  begin
    AdjustReply(offset + 2, result);
  end;
end;

procedure TRcxLink.AdjustReply(offset, length: integer);
var
  src : PByte;
  dst : PByte;
begin
  src := fRxData;
  inc(src, offset);
  dst := fRxData;

  while length > 0 do
  begin
    dst^ := src^;
    inc(dst);
    inc(src);
    if not fFastMode then
      inc(src);
    dec(length);
  end;
end;

procedure TRcxLink.BeginProgress(total: integer);
begin
  fDownloadTotal := total;
  fDownloadSoFar := 0;
  DoDownloadStart;
end;

procedure TRcxLink.BuildTxData(data: PByte; length: integer;
  duplicateReduction: boolean);
var
  i : integer;
  b, checksum : Byte;
  ptr : PByte;
begin
  checksum := 0;
  ptr := fTxData;

  if fTarget = rtCybermaster then
  begin
    // CM sync pattern
    ptr^ := $fe;
    inc(ptr);
    ptr^ := $0;
    inc(ptr);
    ptr^ := $0;
    inc(ptr);
    ptr^ := $FF;
    inc(ptr);
  end
  else if fFastMode then
  begin
    ptr^ := $FF;
    inc(ptr);
  end
  else
  begin
    // RCX sync pattern
    ptr^ := $55;
    inc(ptr);
    ptr^ := $FF;
    inc(ptr);
    ptr^ := $0;
    inc(ptr);
  end;

  // interleaved data & inverse data
  for i := 0 to length - 1 do
  begin
    b := data^;
    inc(data);
    if i = 0 then
    begin
      if duplicateReduction and (b = fTxLastCommand) then
        b := b xor 8;
      fTxLastCommand := b;
    end;
    ptr^ := b;
    inc(ptr);
    if not fFastMode then
    begin
      ptr^ := Byte(not b);
      inc(ptr);
    end;
    checksum := checksum + b;
  end;
  // checksum
  ptr^ := checksum;
  inc(ptr);
  if not fFastMode then
  begin
    ptr^ := Byte(not checksum);
    inc(ptr);
  end;
  fTxLength := (DWord(ptr) - DWord(fTxData)) div sizeof(Byte);
end;

function TRcxLink.Close: boolean;
begin
  fSynced := False;
  NQCSerial.Close;
  result := True;
end;

constructor TRcxLink.Create;
begin
  inherited Create;
  fIgnoreReply := False;
  fLog := '';
  fTxData := AllocMem(kMaxTxData * sizeof(Byte));
  fRxData := AllocMem(kMaxRxData * sizeof(Byte));
  fVerbose := False;
  fSynced  := False;
  fTxLastCommand := 0;
  SetTarget(rtRCX);
  fPortNum := 1;
  fFastMode := False;
end;

destructor TRcxLink.Destroy;
begin
  Close;
  FreeMem(fTxData);
  FreeMem(fRxData);
  if assigned(fSerial) then fSerial.Free;
  inherited Destroy;
end;

function TRcxLink.Download(const data: PByte; length,
  chunk: integer): integer;
var
  cmd : TRcxCmd;
  seq : Word;
  remain, n, timeout : integer;
  tmpD : PByte;
begin
  cmd := TRcxCmd.Create;
  try
    tmpD := data;
    remain := length;
    seq := 1;
    while remain > 0 do
    begin
      if remain <= chunk then
      begin
        seq := 0;
        n := remain;
      end
      else
      begin
        n := chunk;
      end;

      timeout := 0;
      if FastMode then timeout := 100;
      result := Send(cmd.MakeDownload(seq, tmpD, Byte(n)), true, timeout);
      inc(seq);
      if result < 0 then Exit;

      dec(remain, n);
      inc(tmpD, n);
      if not IncrementProgress(n) then
      begin
        result := kRCX_AbortError;
        Exit;
      end;
    end;
    result := kRCX_OK;
  finally
    cmd.Free;
  end;
end;

function TRcxLink.DownloadProgress(soFar, total: integer): boolean;
begin
  result := true;
end;

function TRcxLink.FindReply(var offset: integer): integer;
var
  tmpRxD : PByte;
begin
  offset := 0;
  while true do
  begin
    tmpRxD := fRxData;
    inc(tmpRxD, offset);
    if fTarget = rtCybermaster then
      result := FindCMSync(tmpRxD, fRxLength - offset)
    else
      result := FindRCXSync(tmpRxD, fRxLength - offset);

    if result = 0 then
    begin
      result := kRCX_ReplyError;
      Exit;
    end;

    inc(offset, result);
    inc(tmpRxD, result);
    if fFastMode then
      result := VerifyFastReply(tmpRxD, fRxLength - offset, fTxLastCommand)
    else
      result := VerifyReply(tmpRxD, fRxLength - offset, fTxLastCommand);

    if result >= 0 then Exit;
  end;
end;

function TRcxLink.GetBatteryLevel: integer;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if RCX_ERROR(result) then Exit;

    if fTarget = rtScout then
    begin
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, $3a, $01, $01));
      if result <> 1 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := Integer(GetReplyByte(0)) * 109;
    end
    else
    begin
      result := Send(cmd.SetVal(kRCX_BatteryLevelOp));
      if result <> 2 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := Integer(GetReplyByte(0)) + (Integer(GetReplyByte(1)) shl 8);
    end;
  finally
    cmd.Free;
  end;
end;

function TRcxLink.GetReply(data: PByte; maxLength: integer): integer;
var
  length : integer;
  src : PByte;
begin
  if fResult < 0 then
  begin
    result := fResult;
    Exit;
  end;

  src := fRxData;

  length := 0;
  while length < maxLength do
  begin
    if length >= fResult then Break;
    data^ := src^;
    inc(data);
    inc(src);
    inc(length);
  end;

  result := length;
end;

function TRcxLink.GetReplyByte(index: integer): Byte;
var
  pb : PByte;
begin
  pb := fRxData;
  inc(pb, index);
  result := pb^;
end;

function TRcxLink.GetValue(value: integer): integer;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if RCX_ERROR(result) then Exit;

    result := Send(cmd.MakeRead(value));
    if RCX_ERROR(result) then Exit;
    if result <> 2 then
    begin
      result := kRCX_ReplyError;
      Exit;
    end;

    result := Integer(GetReplyByte(0)) + (Integer(GetReplyByte(1)) shl 8);
  finally
    cmd.Free;
  end;
end;

function TRcxLink.GetVerbose: boolean;
begin
  result := fVerbose;
end;

function TRcxLink.GetVersion(var rom, ram: Cardinal): integer;
var
  cmd : TRcxCmd;
  reply : array[0..7] of Byte;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if RCX_ERROR(result) then Exit;

    result := Send(cmd.MakeUnlock());
    if RCX_ERROR(result) then Exit;

    if result <> 8 then
    begin
      result := kRCX_ReplyError;
      Exit;
    end;

    GetReply(@reply[0],8);

    rom := (Cardinal(reply[0]) shl 24) or (Cardinal(reply[1]) shl 16) or
           (Cardinal(reply[2]) shl 8) or Cardinal(reply[3]);
    ram := (Cardinal(reply[4]) shl 24) or (Cardinal(reply[5]) shl 16) or
           (Cardinal(reply[6]) shl 8) or Cardinal(reply[7]);

    result := kRCX_OK;
  finally
    cmd.Free;
  end;
end;

function TRcxLink.IncrementProgress(delta: integer): boolean;
begin
  inc(fDownloadSoFar, delta);
  if fDownloadTotal <> 0 then
    result := DownloadProgress(fDownloadSoFar, fDownloadTotal)
  else
    result := true;
end;

function TRcxLink.Open(aPortNum : TPortNum): boolean;
var
  myport : string;
begin
  PortNum := aPortNum;
  myport := PortName;
  Result := NQCSerial.Open(myport);
  if not Result then Exit;

  if fFastMode then
    NQCSerial.SetSpeed(4800)
  else
    NQCSerial.SetSpeed(2400, kPSerial_ParityOdd);

  if fTarget = rtCybermaster then
  begin
    NQCSerial.SetDTR(true);
    NQCSerial.SetRTS(false);
  end;
end;

procedure TRcxLink.ProcessRxByte(b: Byte);
var
  newState : integer;
  reset : boolean;
  tmpRxD : PByte;
begin
  newState := fRxState;
  reset := false;

  case fRxState of
    kIdleState :
      if b = $55 then newState := kSync1State;
    kSync1State :
      if b = $ff then
      	newState := kSync2State
      else
	reset := true;
    kSync2State :
      if b = $00 then
      begin
	fRxLength := 0;
	newState := kDataState;
      end
      else
	reset := true;
    kDataState :
      if fRxLength < kMaxRxData then
      begin
        tmpRxD := fRxData;
        inc(tmpRxD, fRxLength);
        tmpRxD^ := b;
        inc(fRxLength);
      end;
  end;

  if reset then
  begin
    if b = $55 then
      newState := kSync1State
    else
      newState := kIdleState;
  end;
  SetRxState(newState);
end;

function TRcxLink.Receive(echo: boolean): integer;
var
  b : array[0..0] of Byte;
  pb : PByte;
begin
  if fRxState = kReplyState then
    SetRxState(kIdleState);

  pb := @b[0];
  while true do
  begin
    if NQCSerial.Read(pb, 1) <> 1 then
    begin
      if fRxState = kDataState then
      	result := ValidateRxData
      else
	result := 0;

      SetRxState(kIdleState);
      Exit;
    end;

    if echo then
      NQCSerial.Write(@b[0], 1);

    ProcessRxByte(b[0]);
  end;
end;

function TRcxLink.Send(cmd: TRcxCmd; retry: boolean; timeout: integer): integer;
begin
  result := Send(cmd.GetBody, cmd.GetLength, retry, timeout);
end;

function TRcxLink.Send(data: PByte; length: integer; retry: boolean;
  timeout: integer): integer;
var
  newTimeout, tries, originalTimeout, i : integer;
  bDynamic : boolean;
begin
  if length > kMaxCmdLength then
  begin
    result := kRCX_RequestError;
    Exit;
  end;

  // format the command
  BuildTxData(data, length, retry);

  // determine reply length
  fRxExpected := PredictReplyLength(data, length);

  // try sending
  if retry then
    tries := kDefaultRetryCount
  else
    tries := 1;
  originalTimeout := fRxTimeout;
  bDynamic := (timeout = 0) and fDynamicTimeout and not fPredictive;

  for i := 0 to tries - 1 do
  begin
    if timeout > 0 then
      fResult := _SendFromTxBuffer(timeout)
    else
      fResult := _SendFromTxBuffer(fRxTimeout);

    if not RCX_ERROR(fResult) then
    begin
      // success
      if i = 0 then
      begin
        // worked on first try, lets see if we can go faster next time
        if bDynamic and (fRxTimeout > kMinTimeout) then
              begin
          newTimeout := fRxTimeout - Trunc(fRxTimeout / 10);
          if newTimeout < kMinTimeout then
            newTimeout := kMinTimeout;
          fRxTimeout := newTimeout;
        end;
      end;
      result := fResult;
      Exit;
    end;

    // only the second kRCX_IREchoError is catastrophic
    // this is somewhat of a hack - I really should keep track
    // of the echo, but for now this makes sure that a serial
    // level failure on a single packet doesn't kill the entire
    // send
    if (fResult = kRCX_IREchoError) and (i > 0) then
      break;

    if bDynamic and (i > 0) then
    begin
      // increase retry time
      fRxTimeout := fRxTimeout * 2;
      if fRxTimeout > kMaxTimeout then
        fRxTimeout := kMaxTimeout;
    end;
  end;

  // retries exceeded, restore original timeout and lose the sync
  if retry then
  begin
    if bDynamic then
      fRxTimeout := originalTimeout;
    fSynced := false;
  end;
  result := fResult;
end;

function TRcxLink.Send1(data: PByte; length: integer): integer;
begin
  if length > kMaxCmdLength then
  begin
    result := kRCX_RequestError;
    Exit;
  end;

  // format the command
  BuildTxData(data, length, false);
  NQCSerial.Write(fTxData, fTxLength);
  result := kRCX_OK;
end;

function TRcxLink.SendFirmware(const data: PByte; length, start: integer;
  fast: boolean): integer;
var
  wasPredictive : boolean;
begin
  wasPredictive := fPredictive;
  if fast then
  begin
    // send the nub first
    result := _SendFirmware(@rcxnub[0], sizeof(rcxnub), kNubStart, false);
    if RCX_ERROR(result) then Exit;

    // switch to fast mode
    fFastMode := true;
    fPredictive := false;
    NQCSerial.SetSpeed(4800);
  end;

  result := _SendFirmware(data, length, start, true);

  if fast then
  begin
    fFastMode := false;
    NQCSerial.SetSpeed(2400, kPSerial_ParityOdd);
    fPredictive := wasPredictive;
  end;
end;

function TRcxLink.SendFragment(ftype: TRcxFragment; taskNumber: Byte;
  const data: PByte; length, total: integer): integer;
var
  cmd : TRcxCmd;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync;
    if RCX_ERROR(result) then Exit;

    result := Send(cmd.MakeBegin(ftype, taskNumber, Word(length)));
    if RCX_ERROR(result) then Exit;

    // make sure we have enough room
    if (result <> 1) or (GetReplyByte(0) <> 0) then
    begin
      result := kRCX_MemFullError;
      Exit;
    end;

    if total = 0 then
      total := length;

    if total > 0 then
      BeginProgress(total);

    result := Download(data, length, kFragmentChunk);
  finally
    cmd.Free;
  end;
end;

procedure TRcxLink.SetRxState(state: integer);
begin
  if fRxState = state then Exit;

  if state = kIdleState then
    NQCSerial.SetTimeout(200)
  else if fRxState = kIdleState then
    NQCSerial.SetTimeout(30);

  fRxState := state;
end;

procedure TRcxLink.SetRxTimeout(t: integer);
begin
  if fRxTimeout <> t then
  begin
    fRxTimeout := t;
    fDynamicTimeout := false;
  end;
end;

procedure TRcxLink.SetVerbose(b: boolean);
begin
  fVerbose := b;
end;

function TRcxLink.Sync(bScoutPower : boolean = true) : integer;
var
  cmd : TRcxCmd;
begin
  if fSynced then begin
    result := kRCX_OK;
    Exit;
  end;
  cmd := TRcxCmd.Create;
  try
    // always start with a ping
    result := Send(cmd.MakePing());
    // if error, try a second time
    if result = kRCX_ReplyError then
      result := Send(cmd);
    if RCX_ERROR(result) then Exit;

    // cybermaster requires an unlock also
    if fTarget = rtCybermaster then
    begin
      result := Send(cmd.MakeUnlockCM());
      if RCX_ERROR(result) then
        Exit;
    end
    else if fTarget = rtScout then
    begin
      result := Send(cmd.MakeBoot());
      if RCX_ERROR(result) then Exit;

      if bScoutPower then
      begin
        result := Send(cmd.SetVal(kRCX_ScoutOp, $80));
        if RCX_ERROR(result) then Exit;
      end;
    end;

    fSynced := true;
    result := kRCX_OK;
  finally
    cmd.Free;
  end;
end;

function TRcxLink.ValidateRxData: integer;
var
  count, i : integer;
  d, checksum : Byte;
  tmpRxD : PByte;
begin
  // validate complements
  count := fRxLength div 2;
  i := 0;
  while i < count do
  begin
    inc(i);
    tmpRxD := fRxData;
    inc(tmpRxD, 2*i);
    d := tmpRxD^;
    inc(tmpRxD);
    if (d + tmpRxD^) <> $ff then
    begin
      result := 0;
      Exit;
    end;
    tmpRxD := fRxData;
    inc(tmpRxD, i);
    tmpRxD^ := d;
  end;

  // compute checksum
  dec(count);
  checksum := 0;
  tmpRxD := fRxData;
  for i := 0 to count - 1 do
  begin
    checksum := checksum + tmpRxD^;
    inc(tmpRxD);
  end;

  tmpRxD := fRxData;
  inc(tmpRxD, count);
  if checksum <> tmpRxD^ then
  begin
    result := 0;
    Exit;
  end;
  result := count;
end;

procedure TRcxLink.SetTarget(target: TRCXType);
begin
  fTarget := target;
  fPredictive := true;
  fDynamicTimeout := true;
  if fPredictive then
    fRxTimeout := kMaxTimeout
  else
    fRxTimeout := kMinTimeout;
end;

procedure TRcxLink.WriteToLog(s: string);
begin
  fLog := fLog + s;
end;

procedure TRcxLink.ClearLog;
begin
  fLog := '';
end;

procedure TRcxLink.DumpData(ptr: PByte; length: integer);
var
  i : integer;
  s : string;
begin
  s := '';
  for i := 0 to length - 1 do
  begin
    s := s + Format('%2x', [ptr^]);
    inc(ptr);
  end;
  WriteToLog(s + #13#10);
end;

procedure TRcxLink.DoDownloadDone;
begin
  if Assigned(fOnDownloadDone) then
    fOnDownloadDone(self);
end;

procedure TRcxLink.DoDownloadStart;
begin
  if Assigned(fOnDownloadStart) then
    fOnDownloadStart(self);
end;

procedure TRcxLink.DoDownloadStatus(cur, total : Integer; var Abort: boolean);
begin
  Abort := False;
  if assigned(fOnDownloadStatus) then
    fOnDownloadStatus(self, cur, total, Abort);
end;

function TRcxLink.GetMemMap(aMemMap: TStrings): integer;
const
  K_RCX_LEN = 188;
  K_CYBER_LEN = 20;
  K_SCOUT_LEN = 22;
var
  cmd : TRcxCmd;
  b1, b2 : Byte;
  i, val : integer;
begin
  aMemMap.Clear;
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if RCX_ERROR(result) then Exit;

    if fTarget = rtScout then
    begin
      // scout memory map starts at $0233 and it is 22 bytes long
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, $33, $02, K_SCOUT_LEN));
      if result <> K_SCOUT_LEN then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      i := 0;
      while i < result do
      begin
        b1  := GetReplyByte(i);
        b2  := GetReplyByte(i+1);
        val := (Cardinal(b2) shl 8) or Cardinal(b1);
        aMemMap.Add(IntToStr(val));
        inc(i, 2);
      end;
    end
    else
    begin
      result := Send(cmd.SetVal(kRCX_MemMapOp));

      if fTarget = rtCybermaster then
      begin
        if result <> K_CYBER_LEN then
        begin
          result := kRCX_ReplyError;
          Exit;
        end;
      end
      else
      begin
        if result <> K_RCX_LEN then
        begin
          result := kRCX_ReplyError;
          Exit;
        end;
      end;
      i := 0;
      while i < result do
      begin
        b1  := GetReplyByte(i);
        b2  := GetReplyByte(i+1);
        val := (Cardinal(b1) shl 8) or Cardinal(b2) - 32768;
        aMemMap.Add(IntToStr(val));
        inc(i, 2);
      end;
    end;
    result := kRCX_OK;
  finally
    cmd.Free;
  end;
end;

procedure TRcxLink.SetIgnoreReply(const Value: boolean);
begin
  fIgnoreReply := Value;
end;

function TRcxLink.GetOutputStatus(aOut: Byte): integer;
var
  cmd : TRcxCmd;
begin
  if fTarget = rtScout then
  begin
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;
      // aOut = 0, 1, 2.  For the scout the motor state is located at memory
      // addresses 144, 145, and 146.
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, $44 + aOut, $01, $01));
      if result <> 1 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := GetReplyByte(0);
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_OutputStatusType, aOut);
  end;
end;

function TRcxLink.GetVariableValue(aVar: Byte): Integer;
var
  cmd : TRcxCmd;
  addr : Word;
begin
  if fTarget = rtScout then
  begin
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;

      // global variables start at ADD_GV and go up from there by Two's
      // aOut = 0, 1, 2.  For the scout the motor state is located at memory
      // addresses 144, 145, and 146.
      addr := ADD_GV + aVar*2;
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(addr), Hi(addr), $02));
      if result <> 2 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := SmallInt(Cardinal(GetReplyByte(1) shl 8) + GetReplyByte(0));
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_VariableType, aVar);
  end;
end;

function TRcxLink.GetInputValue(aIn: Byte): Integer;
var
  cmd : TRcxCmd;
  addr : Word;
begin
  if fTarget = rtScout then
  begin
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;
      // inputs start at ADD_IO and go up from there by Two's
      addr := ADD_IO + aIn*2;
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(addr), Hi(addr), $02));
      if result <> 2 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := SmallInt(Cardinal(GetReplyByte(1) shl 8) + GetReplyByte(0));
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_InputValueType, aIn);
  end;
end;

function TRcxLink.GetMessageValue(aNum: Byte) : Integer;
var
  cmd : TRcxCmd;
  addr : Word;
begin
  if fTarget = rtScout then
  begin
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;
      // PB Message is at ADD_MESSAGE (aNum is ignored)
      addr := ADD_MESSAGE;
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(addr), Hi(addr), $01));
      if result <> 1 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := GetReplyByte(0);
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_MessageType, aNum);
  end;
end;

function TRcxLink.GetTimerValue(aNum: Byte): Integer;
var
  cmd : TRcxCmd;
  addr : Word;
begin
  if fTarget = rtScout then
  begin
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;
      // counters are at ADD_TMR and go up by 2
      addr := ADD_TMR + aNum*2;
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(addr), Hi(addr), $02));
      if result <> 2 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := SmallInt(Cardinal(GetReplyByte(1) shl 8) + GetReplyByte(0));
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_TimerType, aNum);
  end;
end;

function TRcxLink.GetCounterValue(aNum: Byte): Integer;
var
  cmd : TRcxCmd;
  addr : Word;
begin
  if fTarget = rtScout then
  begin
    if aNum > 1 then
    begin
      result := 0;
      Exit;
    end;
    cmd := TRcxCmd.Create;
    try
      result := Sync();
      if RCX_ERROR(result) then Exit;
      // counters are at ADD_CTR and go up by 2
      addr := ADD_CTR + aNum*2;
      result := Send(cmd.SetVal(kRCX_PollMemoryOp, Lo(addr), Hi(addr), $02));
      if result <> 2 then
      begin
        result := kRCX_ReplyError;
        Exit;
      end;
      result := SmallInt(Cardinal(GetReplyByte(1) shl 8) + GetReplyByte(0));
    finally
      cmd.Free;
    end;
  end
  else
  begin
    result := Poll(kRCX_CounterType, aNum);
  end;
end;

function TRcxLink.Poll(aSrc, aNum: integer): integer;
var
  cmd : TRcxCmd;
  val, cnt : integer;
begin
  cmd := TRcxCmd.Create;
  try
    result := Sync();
    if RCX_ERROR(result) then Exit;

    val := integer((aSrc shl 16) or (aNum and $FFFF));
    cnt := Send(cmd.MakeRead(val));
    if cnt <> 2 then
    begin
      result := -1;
      Exit;
    end;
    result := SmallInt(Cardinal(GetReplyByte(1) shl 8) + GetReplyByte(0));
  finally
    cmd.Free;
  end;
end;

procedure TRcxLink.SetFastMode(const Value: boolean);
begin
  Close;
  fFastMode := Value;
end;

function TRcxLink.GetNQCSerial: TNQCSerial;
begin
  if not Assigned(fSerial) then begin
    if PortIsUSB(GetPortNum) then
      fSerial := TNQCUsbTower.Create
    else
      fSerial := TNQCSerial.Create;
  end;
  result := fSerial;
end;

function TRcxLink.GetPortName: string;
begin
  if PortIsUSB(GetPortNum) then
    Result := USB_TOWER + IntToStr(GetPortNum - MAX_COMPORT)
  else
    Result := COM_PORT + IntToStr(GetPortNum);
end;

function TRcxLink.GetPortNum: TPortNum;
begin
  Result := fPortNum;
end;

procedure TRcxLink.SetPortNum(const Value: TPortNum);
var
  tmp : TNQCSerial;
begin
  if Value <> fPortNum then begin
    fPortNum := Value;
    tmp := fSerial;
    fSerial := nil;
    tmp.Free;
  end;
end;

procedure TRcxLink.DoOpenStateChanged;
begin
  if Assigned(FOnOpenStateChanged) then
    FOnOpenStateChanged(self);
end;

{ TAutoLink }

function TAutoLink.Close: boolean;
begin
  if fOpen then
  begin
    inherited Close;
    fOpen := False;
    DoOpenStateChanged;
  end;
  Result := True;
end;

constructor TAutoLink.Create;
begin
  inherited Create;
  fOpen := False;
end;

function TAutoLink.DownloadProgress(soFar, total: integer): boolean;
var
  abort : boolean;
begin
  result := inherited DownloadProgress(soFar, total);
  if result then
  begin
    DoDownloadStatus(soFar, total, abort);
    result := not abort;
  end;
  if abort or (soFar >= total) then
    DoDownloadDone;
end;

function TAutoLink.Open : boolean;
begin
  if not fOpen then
  begin
    result := inherited Open(PortNum);
    fOpen := result;
    if fOpen then
      DoOpenStateChanged;
  end
  else
    result := True;
end;

function TAutoLink.Send(cmd: TRcxCmd; bRetry: boolean): integer;
var
  val : boolean;
begin
  result := kRCX_OpenSerialError;
  val := Open;
  if val then
  begin
    if bRetry then
    begin
      result := Sync;
      val := RCX_ERROR(result);
      if val then Exit;
    end;
    result := inherited Send(cmd, bRetry);
//    if not bRetry then
//      result := kRCX_OK;
  end;
end;

procedure TAutoLink.SetPortNum(const Value: TPortNum);
var
  tmp : TNQCSerial;
begin
  if Value <> fPortNum then begin
    fPortNum := Value;
    tmp := fSerial;
    fSerial := nil;
    tmp.Free;
    Close;
  end;
end;

end.
